@@ -27,6 +27,7 @@ extern "C" {
  * Exit status bitmask values.
  */
 #define PV_ERROREXIT_REMOTE_OR_PID 1
+#define PV_ERROREXIT_SAF	   1	/* store and forward error */
 #define PV_ERROREXIT_ACCESS        2
 #define PV_ERROREXIT_OUROBOROS     4
 #define PV_ERROREXIT_TRANSITION    8
@@ -78,6 +79,13 @@ extern unsigned int pv_getnum_count(const char *, bool);
  */
 extern bool pv_getnum_check(const char *, pv_numtype);
 
+/*
+ * Return an integer representing the first amount as a percentage of the
+ * second total, i.e. 100*amount/total.  If the second value, the total, is
+ * zero or less, return 0.
+ */
+extern int pv_percentage(off_t, const off_t);
+
 /*
  * String handling wrappers.
  */
@@ -156,6 +164,11 @@ void pv_nanosleep(long long);
  */
 extern /*@null@*/ /*@only@*/ pvstate_t pv_state_alloc(const char *);
 
+/*
+ * Clear the calculated parts of a state structure.
+ */
+extern void pv_state_reset(pvstate_t state);
+
 /*
  * Set the formatting string, given a set of old-style formatting options.
  */
@@ -172,9 +185,11 @@ extern void pv_state_set_format(pvstate_t state, bool progress,
  */
 extern void pv_state_force_set(pvstate_t, bool);
 extern void pv_state_cursor_set(pvstate_t, bool);
+extern void pv_state_show_stats_set(pvstate_t, bool);
 extern void pv_state_numeric_set(pvstate_t, bool);
 extern void pv_state_wait_set(pvstate_t, bool);
 extern void pv_state_delay_start_set(pvstate_t, double);
+extern void pv_state_rate_gauge_set(pvstate_t, bool);
 extern void pv_state_linemode_set(pvstate_t, bool);
 extern void pv_state_bits_set(pvstate_t, bool);
 extern void pv_state_decimal_units_set(pvstate_t, bool);
