@@ -66,7 +66,7 @@ bool pv_in_foreground(void)
 	pid_t tty_process_group;
 
 	if (0 == isatty(STDERR_FILENO)) {
-		debug("%s: true: %s", "pv_in_foreground", "not a tty");
+		debug("true: %s", "not a tty");
 		return true;
 	}
 
@@ -76,22 +76,77 @@ bool pv_in_foreground(void)
 	/*@+type@ */
 
 	if (tty_process_group == -1 && errno == ENOTTY) {
-		debug("%s: true: %s", "pv_in_foreground", "tty_process_group is -1, errno is ENOTTY");
+		debug("true: %s", "tty_process_group is -1, errno is ENOTTY");
 		return true;
 	}
 
 	if (our_process_group == tty_process_group) {
-		debug("%s: true: %s", "pv_in_foreground", "our_process_group == tty_process_group");
+		debug("true: %s (%d)", "our_process_group == tty_process_group", our_process_group);
 		return true;
 	}
 
-	debug("%s: false: our_process_group=%d, tty_process_group=%d",
-	      "pv_in_foreground", our_process_group, tty_process_group);
+	/*
+	 * If the terminal process group ID doesn't match our own, assume
+	 * we're in the background.
+	 */
+	debug("false: our_process_group=%d, tty_process_group=%d", our_process_group, tty_process_group);
 
 	return false;
 }
 
 
+/*
+ * Write the given buffer to the given file descriptor, retrying until all
+ * bytes have been written or an error has occurred.
+ */
+void pv_write_retry(int fd, const char *buf, size_t count)
+{
+	while (count > 0) {
+		ssize_t nwritten;
+
+		nwritten = write(fd, buf, count);
+
+		if (nwritten < 0) {
+			if ((EINTR == errno) || (EAGAIN == errno)) {
+				continue;
+			}
+			return;
+		}
+		if (nwritten < 1)
+			return;
+
+		count -= nwritten;
+		buf += nwritten;
+	}
+}
+
+
+/*
+ * Write the given buffer to the terminal, like pv_write_retry(), unless
+ * stderr is suspended.
+ */
+void pv_tty_write(pvstate_t state, const char *buf, size_t count)
+{
+	while (0 == state->flag.suspend_stderr && count > 0) {
+		ssize_t nwritten;
+
+		nwritten = write(STDERR_FILENO, buf, count);
+
+		if (nwritten < 0) {
+			if ((EINTR == errno) || (EAGAIN == errno)) {
+				continue;
+			}
+			return;
+		}
+		if (nwritten < 1)
+			return;
+
+		count -= nwritten;
+		buf += nwritten;
+	}
+}
+
+
 /*
  * Fill in *width and *height with the current terminal size,
  * if possible.
@@ -113,21 +168,6 @@ void pv_screensize(unsigned int *width, unsigned int *height)
 }
 
 
-/*
- * Calculate the percentage transferred so far and return it.
- */
-static int pv__calc_percentage(off_t so_far, const off_t total)
-{
-	if (total < 1)
-		return 0;
-
-	so_far *= 100;
-	so_far /= total;
-
-	return (int) so_far;
-}
-
-
 /*
  * Given how many bytes have been transferred, the total byte count to
  * transfer, and the current average transfer rate, return the estimated
@@ -422,15 +462,15 @@ static void pv__format_init(pvstate_t state)
 	 * with a particular length.
 	 *
 	 * A type other than PV_COMPONENT_STRING indicates that the segment
-	 * is a string updated by pv__format(), whose contents will be in
+	 * is a string updated by pv_format(), whose contents will be in
 	 * component[type].
 	 *
-	 * In pv__format(), the content of a PV_COMPONENT_PROGRESS component
+	 * In pv_format(), the content of a PV_COMPONENT_PROGRESS component
 	 * is calculated after first populating all the other components
 	 * referenced by the format segments.
 	 *
-	 * Then, pv_format() generates the output string by sticking all of
-	 * these segments together.
+	 * Then, that function generates the output string by sticking all
+	 * of these segments together.
 	 */
 	segment = 0;
 	for (strpos = 0; formatstr[strpos] != '\0' && segment < PV_FORMAT_ARRAY_MAX; strpos++, segment++) {
@@ -552,6 +592,7 @@ static void pv__format_init(pvstate_t state)
 	}
 }
 
+
 /*
  * Return the original value x so that it has been clamped between
  * [min..max]
@@ -561,103 +602,41 @@ static long bound_long(long x, long min, long max)
 	return x < min ? min : x > max ? max : x;
 }
 
-/*
- * Update the current average rate, using a ring buffer of past transfer
- * positions - if this is the first entry, use the provided instantaneous
- * rate, otherwise calulate the average rate from the difference between the
- * current position + elapsed time pair, and the oldest pair in the buffer.
- */
-static void pv__update_average_rate_history(pvstate_t state, off_t total_bytes, long double elapsed_sec,
-					    long double rate)
-{
-	size_t first = state->display.history_first;
-	size_t last = state->display.history_last;
-	long double last_elapsed;
-
-	if (NULL == state->display.history)
-		return;
-
-	last_elapsed = state->display.history[last].elapsed_sec;
-
-	/*
-	 * Do nothing if this is not the first call but not enough time has
-	 * elapsed since the previous call yet.
-	 */
-	if ((last_elapsed > 0.0)
-	    && (elapsed_sec < (last_elapsed + state->display.history_interval)))
-		return;
-
-	/*
-	 * If this is not the first call, add a new entry to the circular
-	 * buffer.
-	 */
-	if (last_elapsed > 0.0) {
-		size_t len = state->display.history_len;
-		last = (last + 1) % len;
-		state->display.history_last = last;
-		if (last == first) {
-			first = (first + 1) % len;
-			state->display.history_first = first;
-		}
-	}
-
-	state->display.history[last].elapsed_sec = elapsed_sec;
-	state->display.history[last].total_bytes = total_bytes;
-
-	if (first == last) {
-		state->display.current_avg_rate = rate;
-	} else {
-		off_t bytes = (state->display.history[last].total_bytes - state->display.history[first].total_bytes);
-		long double sec =
-		    (state->display.history[last].elapsed_sec - state->display.history[first].elapsed_sec);
-		state->display.current_avg_rate = (long double) bytes / sec;
-	}
-}
 
 /*
  * Update state->display.display_buffer with status information formatted
- * according to the state held within the given structure, where
- * "elapsed_sec" is the seconds elapsed since the transfer started,
- * "bytes_since_last" is the number of bytes transferred since the last
- * update, and "total_bytes" is the total number of bytes transferred so
- * far.
+ * according to the state held within the given structure.
  *
- * If "bytes_since_last" is negative, this is the final update so the rate
- * is given as an an average over the whole transfer; otherwise the current
- * rate is shown.
- *
- * In line mode, "bytes_since_last" and "total_bytes" are in lines, not bytes.
+ * If "final" is true, this is the final update so the rate is given as an
+ * an average over the whole transfer; otherwise the current rate is shown.
  *
  * Returns true if the display buffer can be used, false if not.
  *
  * When returning true, this function will have also set
  * state->display.display_string_len to the length of the string in
  * state->display.display_buffer, in bytes.
- *
- * If "total_bytes" is negative, then free the display buffer and return
- * false.
  */
-static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_since_last, off_t total_bytes)
+bool pv_format(pvstate_t state, bool final)
 {
-	long double time_since_last, rate, average_rate;
 	long eta;
 	int static_portion_size;
 	pv_display_component component_type;
 	size_t segment;
 	size_t new_display_string_len;
 	const char *formatstr;
+	pv__transfercount_t count_type;
 
-	/* Quick sanity check - state must exist. */
+	/* Quick safety check - state must exist. */
 	if (NULL == state)
 		return false;
 
-	/* Negative total transfer - free memory and return false. */
-	if (total_bytes < 0) {
-		if (NULL != state->display.display_buffer)
-			free(state->display.display_buffer);
-		state->display.display_buffer = NULL;
-		state->display.display_buffer_size = 0;
-		return false;
+	/*
+	 * If the display options need reparsing, do so to generate new
+	 * formatting parameters.
+	 */
+	if (0 != state->flag.reparse_display) {
+		pv__format_init(state);
+		state->flag.reparse_display = 0;
 	}
 
 	/* The format string is needed for the static segments. */
@@ -665,63 +644,12 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 	if (NULL == formatstr)
 		return false;
 
-	/*
-	 * In case the time since the last update is very small, we keep
-	 * track of amount transferred since the last update, and just keep
-	 * adding to that until a reasonable amount of time has passed to
-	 * avoid rate spikes or division by zero.
-	 */
-	time_since_last = elapsed_sec - state->display.prev_elapsed_sec;
-	if (time_since_last <= 0.01) {
-		rate = state->display.prev_rate;
-		state->display.prev_trans += bytes_since_last;
-	} else {
-		rate = ((long double) bytes_since_last + state->display.prev_trans) / time_since_last;
-		state->display.prev_elapsed_sec = elapsed_sec;
-		state->display.prev_trans = 0;
-	}
-	state->display.prev_rate = rate;
-
-	/* Update history and current average rate for ETA. */
-	pv__update_average_rate_history(state, total_bytes, elapsed_sec, rate);
-	average_rate = state->display.current_avg_rate;
-
-	/*
-	 * If this is the final update at the end of the transfer, we
-	 * recalculate the rate - and the average rate - across the whole
-	 * period of the transfer.
-	 */
-	if (bytes_since_last < 0) {
-		/* Sanity check to avoid division by zero */
-		if (elapsed_sec < 0.000001)
-			elapsed_sec = 0.000001;
-		average_rate =
-		    (((long double) total_bytes) -
-		     ((long double) state->display.initial_offset)) / (long double) elapsed_sec;
-		rate = average_rate;
-	}
-
-	if (state->control.size <= 0) {
-		/*
-		 * If we don't know the total size of the incoming data,
-		 * then for a percentage, we gradually increase the
-		 * percentage completion as data arrives, to a maximum of
-		 * 200, then reset it - we use this if we can't calculate
-		 * it, so that the numeric percentage output will go
-		 * 0%-100%, 100%-0%, 0%-100%, and so on.
-		 */
-		if (rate > 0)
-			state->display.percentage += 2;
-		if (state->display.percentage > 199)
-			state->display.percentage = 0;
-	} else if (state->control.numeric || state->display.component[PV_COMPONENT_PROGRESS].required) {
-		/*
-		 * If we do know the total size, and we're going to show
-		 * the percentage (numeric mode or a progress bar),
-		 * calculate the percentage completion.
-		 */
-		state->display.percentage = pv__calc_percentage(total_bytes, state->control.size);
-	}
+	/* Determine the type of thing being counted for transfer, rate, etc. */
+	count_type = PV_TRANSFERCOUNT_BYTES;
+	if (state->control.linemode)
+		count_type = PV_TRANSFERCOUNT_LINES;
+	else if (state->control.decimal_units)
+		count_type = PV_TRANSFERCOUNT_DECBYTES;
 
 	/*
 	 * Reallocate output buffer if width changes.
@@ -759,37 +687,67 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 	}
 
 	/*
-	 * In numeric output mode, our output is just a number.
+	 * In numeric output mode, our output is just the percentage
+	 * completion, as a number by itself.
 	 *
-	 * Patch from Sami Liedes:
-	 * With --timer we prefix the output with the elapsed time.
-	 * With --bytes we output the bytes transferred so far instead
-	 * of the percentage. (Or lines, if --lines was given with --bytes).
+	 * With --timer, we prefix the output with the elapsed time.
+	 *
+	 * With --bytes, we output the bytes transferred so far instead of
+	 * the percentage (or we output the number lines transferred, if
+	 * --lines was given with --bytes).
+	 *
+	 * With --rate was given, we output the current transfer rate
+	 * instead of the percentage.  With --bytes as well, the rate is
+	 * given after the bytes/lines.
 	 */
 	if (state->control.numeric) {
-		char numericprefix[128]; /* flawfinder: ignore - only populated by pv_snprintf(). */
+		char msg_timer[128];	 /* flawfinder: ignore */
+		char msg_bytes[128];	 /* flawfinder: ignore */
+		char msg_rate[128];	 /* flawfinder: ignore */
+		char msg_percent[128];	 /* flawfinder: ignore */
+		bool first_item, show_percentage;
 
-		numericprefix[0] = '\0';
+		/* flawfinder: each buffer is kept safe by pv_snprintf(). */
 
-		if (state->display.component[PV_COMPONENT_TIMER].required)
-			(void) pv_snprintf(numericprefix, sizeof(numericprefix), "%.4Lf ", elapsed_sec);
+		first_item = true;
+		show_percentage = true;
 
-		if (state->display.component[PV_COMPONENT_BYTES].required) {
-			if (state->control.bits) {
-				(void) pv_snprintf(state->display.display_buffer,
-						   state->display.display_buffer_size,
-						   "%.99s%lld\n", numericprefix, (long long) (8 * total_bytes));
-			} else {
-				(void) pv_snprintf(state->display.display_buffer,
-						   state->display.display_buffer_size,
-						   "%.99s%lld\n", numericprefix, (long long) total_bytes);
-			}
-		} else {
-			(void) pv_snprintf(state->display.display_buffer,
-					   state->display.display_buffer_size, "%.99s%ld\n", numericprefix,
-					   (long) (state->display.percentage));
+		msg_timer[0] = '\0';
+		if (state->display.component[PV_COMPONENT_TIMER].required) {
+			(void) pv_snprintf(msg_timer, sizeof(msg_timer), "%s%.4Lf", first_item ? "" : " ",
+					   state->transfer.elapsed_seconds);
+			first_item = false;
 		}
 
+		msg_bytes[0] = '\0';
+		if (state->display.component[PV_COMPONENT_BYTES].required) {
+			(void) pv_snprintf(msg_bytes, sizeof(msg_bytes),
+					   "%s%lld", first_item ? "" : " ",
+					   (long long) ((state->control.bits ? 8 : 1) * state->transfer.transferred));
+			first_item = false;
+			show_percentage = false;
+		}
+
+		msg_rate[0] = '\0';
+		if (state->display.component[PV_COMPONENT_RATE].required) {
+			(void) pv_snprintf(msg_rate, sizeof(msg_rate),
+					   "%s%.4Lf", first_item ? "" : " ",
+					   ((state->control.bits ? 8.0 : 1.0) * state->calc.transfer_rate));
+			first_item = false;
+			show_percentage = false;
+		}
+
+		msg_percent[0] = '\0';
+		if (show_percentage) {
+			(void) pv_snprintf(msg_percent, sizeof(msg_percent),
+					   "%s%d", first_item ? "" : " ", state->calc.percentage);
+			first_item = false;
+		}
+
+		(void) pv_snprintf(state->display.display_buffer,
+				   state->display.display_buffer_size, "%.39s%.39s%.39s%.39s\n", msg_timer, msg_bytes,
+				   msg_rate, msg_percent);
+
 		state->display.display_string_len = strlen(state->display.display_buffer);	/* flawfinder: ignore */
 		/* flawfinder: always \0 terminated by pv_snprintf(). */
 
@@ -807,12 +765,11 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 		char *component_content;
 		size_t component_buf_size;
 		size_t buf_idx;
-		bool show_eta;
+		bool show_fineta;
 		time_t now;
 		time_t then;
 		struct tm *time_ptr;
 		char *time_format;
-		pv__transfercount_t count_type;
 
 		if (!state->display.component[component_type].required)
 			continue;
@@ -833,12 +790,6 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 		component_content[0] = '\0';
 		component_buf_size = PV_SIZEOF_COMPONENT_STR;
 
-		count_type = PV_TRANSFERCOUNT_BYTES;
-		if (state->control.linemode)
-			count_type = PV_TRANSFERCOUNT_LINES;
-		else if (state->control.decimal_units)
-			count_type = PV_TRANSFERCOUNT_DECBYTES;
-
 		switch (component_type) {
 
 		case PV_COMPONENT_STRING:
@@ -853,10 +804,10 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			/*@-mustfreefresh @ */
 			if (state->control.bits && !state->control.linemode) {
 				pv__sizestr(component_content, component_buf_size, "%s",
-					    (long double) total_bytes * 8, "", _("b"), count_type);
+					    (long double) (state->transfer.transferred * 8), "", _("b"), count_type);
 			} else {
 				pv__sizestr(component_content, component_buf_size, "%s",
-					    (long double) total_bytes, "", _("B"), count_type);
+					    (long double) (state->transfer.transferred), "", _("B"), count_type);
 			}
 			/*@+mustfreefresh @ */
 			/* splint: we trust gettext() not to really leak memory. */
@@ -869,26 +820,28 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			 * does mean that the timer will stop at a 100,000 hours,
 			 * but since that's 11 years, it shouldn't be a problem.
 			 */
-			if (elapsed_sec > (long double) 360000000.0L)
-				elapsed_sec = (long double) 360000000.0L;
+			if (state->transfer.elapsed_seconds > (long double) 360000000.0L)
+				state->transfer.elapsed_seconds = (long double) 360000000.0L;
 
 			/*
 			 * If the elapsed time is more than a day, include a day count as
 			 * well as hours, minutes, and seconds.
 			 */
-			if (elapsed_sec > (long double) 86400.0L) {
+			if (state->transfer.elapsed_seconds > (long double) 86400.0L) {
 				(void) pv_snprintf(component_content,
 						   component_buf_size,
 						   "%ld:%02ld:%02ld:%02ld",
-						   ((long) elapsed_sec) / 86400,
-						   (((long) elapsed_sec) / 3600) %
-						   24, (((long) elapsed_sec) / 60) % 60, ((long) elapsed_sec) % 60);
+						   ((long) (state->transfer.elapsed_seconds)) / 86400,
+						   (((long) (state->transfer.elapsed_seconds)) / 3600) %
+						   24, (((long) (state->transfer.elapsed_seconds)) / 60) % 60,
+						   ((long) (state->transfer.elapsed_seconds)) % 60);
 			} else {
 				(void) pv_snprintf(component_content,
 						   component_buf_size,
 						   "%ld:%02ld:%02ld",
-						   ((long) elapsed_sec) / 3600,
-						   (((long) elapsed_sec) / 60) % 60, ((long) elapsed_sec) % 60);
+						   ((long) (state->transfer.elapsed_seconds)) / 3600,
+						   (((long) (state->transfer.elapsed_seconds)) / 60) % 60,
+						   ((long) (state->transfer.elapsed_seconds)) % 60);
 			}
 			break;
 
@@ -897,12 +850,12 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			/*@-mustfreefresh @ */
 			if (state->control.bits && !state->control.linemode) {
 				/* bits per second */
-				pv__sizestr(component_content, component_buf_size, "[%s]", 8 * rate, "", _("b/s"),
-					    count_type);
+				pv__sizestr(component_content, component_buf_size, "[%s]",
+					    8 * state->calc.transfer_rate, "", _("b/s"), count_type);
 			} else {
 				/* bytes or lines per second */
 				pv__sizestr(component_content, component_buf_size,
-					    "[%s]", rate, _("/s"), _("B/s"), count_type);
+					    "[%s]", state->calc.transfer_rate, _("/s"), _("B/s"), count_type);
 			}
 			/*@+mustfreefresh @ *//* splint: see above. */
 			break;
@@ -913,11 +866,12 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			if (state->control.bits && !state->control.linemode) {
 				/* bits per second */
 				pv__sizestr(component_content, component_buf_size,
-					    "[%s]", 8 * average_rate, "", _("b/s"), count_type);
+					    "(%s)", 8 * state->calc.average_rate, "", _("b/s"), count_type);
 			} else {
 				/* bytes or lines per second */
 				pv__sizestr(component_content,
-					    component_buf_size, "[%s]", average_rate, _("/s"), _("B/s"), count_type);
+					    component_buf_size, "(%s)", state->calc.average_rate, _("/s"), _("B/s"),
+					    count_type);
 			}
 			/*@+mustfreefresh @ *//* splint: see above. */
 			break;
@@ -925,9 +879,9 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 		case PV_COMPONENT_ETA:
 			/* Estimated time remaining until completion - if size is known. */
 			eta =
-			    pv__seconds_remaining(((off_t) total_bytes - state->display.initial_offset),
+			    pv__seconds_remaining((state->transfer.transferred - state->display.initial_offset),
 						  state->control.size - state->display.initial_offset,
-						  state->display.current_avg_rate);
+						  state->calc.current_avg_rate);
 
 			/*
 			 * Bounds check, so we don't overrun the suffix buffer. This
@@ -957,7 +911,7 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			 * If this is the final update, show a blank space where the
 			 * ETA used to be.
 			 */
-			if (bytes_since_last < 0) {
+			if (final) {
 				size_t erase_idx;
 				for (erase_idx = 0;
 				     erase_idx < component_buf_size && component_content[erase_idx] != '\0';
@@ -971,18 +925,18 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			/* Estimated time of completion - if size is known. */
 
 			now = time(NULL);
-			show_eta = true;
+			show_fineta = true;
 			time_format = NULL;
 
 			/*
-			 * The ETA may be hidden by a failed ETA string
-			 * generation.
+			 * The completion clock time may be hidden by a
+			 * failed localtime lookup.
 			 */
 
 			eta =
-			    pv__seconds_remaining((off_t) (total_bytes - state->display.initial_offset),
+			    pv__seconds_remaining(state->transfer.transferred - state->display.initial_offset,
 						  state->control.size - state->display.initial_offset,
-						  state->display.current_avg_rate);
+						  state->calc.current_avg_rate);
 
 			/*
 			 * Bounds check, so we don't overrun the suffix buffer. This
@@ -1004,7 +958,7 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			time_ptr = localtime(&then);
 
 			if (NULL == time_ptr) {
-				show_eta = false;
+				show_fineta = false;
 			} else {
 				/* Localtime keeps data stored in a static
 				 * buffer that gets overwritten by time
@@ -1014,7 +968,7 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 				size_t component_content_length;
 
 				/*@-mustfreefresh @ */
-				(void) pv_snprintf(component_content, component_buf_size, "%.16s ", _("ETA"));
+				(void) pv_snprintf(component_content, component_buf_size, "%.16s ", _("FIN"));
 				/*@+mustfreefresh @ *//* splint: see above. */
 				component_content_length = strlen(component_content);	/* flawfinder: ignore */
 				/* flawfinder: always bounded with \0 by pv_snprintf(). */
@@ -1022,7 +976,7 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 						component_buf_size - 1 - component_content_length, time_format, &time);
 			}
 
-			if (!show_eta) {
+			if (!show_fineta) {
 				size_t erase_idx;
 				for (erase_idx = 0;
 				     erase_idx < component_buf_size && component_content[erase_idx] != '\0';
@@ -1043,11 +997,11 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 		case PV_COMPONENT_BUFPERCENT:
 			/* Transfer buffer percentage utilisation. */
 			if (state->transfer.buffer_size > 0) {
-				int pct_used = pv__calc_percentage((off_t)
-								   (state->transfer.read_position -
-								    state->transfer.write_position),
-								   (off_t)
-								   (state->transfer.buffer_size));
+				int pct_used = pv_percentage((off_t)
+							     (state->transfer.read_position -
+							      state->transfer.write_position),
+							     (off_t)
+							     (state->transfer.buffer_size));
 				(void) pv_snprintf(component_content, component_buf_size, "{%3d%%}", pct_used);
 			}
 #ifdef HAVE_SPLICE
@@ -1098,31 +1052,54 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 	if (state->display.component[PV_COMPONENT_PROGRESS].required) {
 		char *component_content;
 		size_t component_buf_size;
-		char pct[16];		 /* flawfinder: ignore - only populated by pv_snprintf(). */
+		char after_bar[32];	 /* flawfinder: ignore - only populated by pv_snprintf(). */
 		int available_width, bar_length, pad_count;
 
 		component_content = state->display.component[PV_COMPONENT_PROGRESS].content;
 		component_content[0] = '\0';
 		component_buf_size = PV_SIZEOF_COMPONENT_STR;
 
-		memset(pct, 0, sizeof(pct));
+		memset(after_bar, 0, sizeof(after_bar));
 
-		/* The opening of the bar area. */
-		(void) pv_snprintf(component_content, component_buf_size, "[");
+		if (state->control.size > 0 || state->control.rate_gauge) {
+			/*
+			 * Known size, or rate gauge mode.  Show a bar, and
+			 * a percentage (size) or max rate (gauge).
+			 */
+			size_t after_bar_width;
+			int bar_percentage;
 
-		if (state->control.size > 0) {
-			/* Known size; show a bar and a percentage. */
-			size_t pct_width;
+			if (state->control.size > 0) {
+				/* Percentage of data transferred. */
+				bar_percentage = state->calc.percentage;
+				(void) pv_snprintf(after_bar, sizeof(after_bar), " %3ld%%", bar_percentage);
+			} else {
+				/* Current rate vs max rate. */
+				bar_percentage = 0;
+				if (state->calc.rate_max > 0) {
+					bar_percentage =
+					    (int) (100.0 * state->calc.transfer_rate / state->calc.rate_max);
+				}
 
-			if (state->display.percentage < 0)
-				state->display.percentage = 0;
-			if (state->display.percentage > 100000)
-				state->display.percentage = 100000;
-			(void) pv_snprintf(pct, sizeof(pct), "%3ld%%", state->display.percentage);
-			pct_width = strlen(pct);	/* flawfinder: ignore */
+				/*@-mustfreefresh @ */
+				if (state->control.bits && !state->control.linemode) {
+					/* bits per second */
+					pv__sizestr(after_bar, sizeof(after_bar), "/%s",
+						    8.0 * state->calc.rate_max, "", _("b/s"), count_type);
+				} else {
+					/* bytes or lines per second */
+					pv__sizestr(after_bar, sizeof(after_bar),
+						    "/%s", state->calc.rate_max, _("/s"), _("B/s"), count_type);
+				}
+				/*@+mustfreefresh @ */
+				/* splint: see above about gettext(). */
+			}
+
+			after_bar_width = strlen(after_bar);	/* flawfinder: ignore */
 			/* flawfinder: always \0-terminated by pv_snprintf() and the earlier memset(). */
 
-			available_width = (int) (state->control.width) - static_portion_size - (int) (pct_width) - 3;
+			available_width =
+			    (int) (state->control.width) - static_portion_size - (int) (after_bar_width) - 2;
 
 			if (available_width < 0)
 				available_width = 0;
@@ -1130,8 +1107,11 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			if (available_width > (int) (component_buf_size) - 16)
 				available_width = (int) (component_buf_size - 16);
 
+			/* The opening of the bar area. */
+			(void) pv_snprintf(component_content, component_buf_size, "[");
+
 			/* The bar portion. */
-			bar_length = (int) ((available_width * state->display.percentage) / 100 - 1);
+			bar_length = (int) ((available_width * bar_percentage) / 100 - 1);
 			for (pad_count = 0; pad_count < bar_length; pad_count++) {
 				if (pad_count < available_width)
 					(void) pv_strlcat(component_content, "=", component_buf_size);
@@ -1149,13 +1129,13 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			}
 
 			/* The closure of the bar area, and the percentage. */
-			(void) pv_strlcat(component_content, "] ", component_buf_size);
-			(void) pv_strlcat(component_content, pct, component_buf_size);
+			(void) pv_strlcat(component_content, "]", component_buf_size);
+			(void) pv_strlcat(component_content, after_bar, component_buf_size);
 
 		} else {
 			/* Unknown size; show a moving indicator. */
 
-			int indicator_position = state->display.percentage;
+			int indicator_position = state->calc.percentage;
 
 			available_width = (int) (state->control.width) - static_portion_size - 5;
 
@@ -1168,15 +1148,18 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 			debug("available_width: %d", available_width);
 
 			/*
-			 * Earlier code in this function sets the percentage
-			 * when the size is unknown to a value that goes 0 -
-			 * 200 and resets, so here we make values above 100
-			 * send the indicator back down again, so it moves
-			 * back and forth.
+			 * Note that pv_calculate_transfer_rate() sets the
+			 * percentage when the size is unknown to a value
+			 * that goes 0 - 200 and resets, so here we make
+			 * values above 100 send the indicator back down
+			 * again, so it moves back and forth.
 			 */
 			if (indicator_position > 100)
 				indicator_position = 200 - indicator_position;
 
+			/* The opening of the bar area. */
+			(void) pv_snprintf(component_content, component_buf_size, "[");
+
 			/* The spaces before the indicator. */
 			for (pad_count = 0; pad_count < (available_width * indicator_position) / 100; pad_count++) {
 				if (pad_count < available_width)
@@ -1288,40 +1271,28 @@ static bool pv__format(pvstate_t state, long double elapsed_sec, off_t bytes_sin
 
 
 /*
- * Output status information on standard error, where "esec" is the seconds
- * elapsed since the transfer started, "sl" is the number of bytes transferred
- * since the last update, and "tot" is the total number of bytes transferred
- * so far.
+ * Output status information on standard error.
  *
- * If "sl" is negative, this is the final update so the rate is given as an
+ * If "final" is true, this is the final update, so the rate is given as an
  * an average over the whole transfer; otherwise the current rate is shown.
- *
- * In line mode, "sl" and "tot" are in lines, not bytes.
  */
-void pv_display(pvstate_t state, long double esec, off_t sl, off_t tot)
+void pv_display(pvstate_t state, bool final)
 {
 	if (NULL == state)
 		return;
 
-	/*
-	 * If the display options need reparsing, do so to generate new
-	 * formatting parameters.
-	 */
-	if (0 != state->flag.reparse_display) {
-		pv__format_init(state);
-		state->flag.reparse_display = 0;
-	}
-
 	pv_sig_checkbg();
 
-	if (!pv__format(state, esec, sl, tot))
+	pv_calculate_transfer_rate(state, final);
+
+	if (!pv_format(state, final))
 		return;
 
 	if (NULL == state->display.display_buffer)
 		return;
 
 	if (state->control.numeric) {
-		pv_write_retry(STDERR_FILENO, state->display.display_buffer, state->display.display_string_len);
+		pv_tty_write(state, state->display.display_buffer, state->display.display_string_len);
 	} else if (state->control.cursor) {
 		if (state->control.force || pv_in_foreground()) {
 			pv_crs_update(state, state->display.display_buffer);
@@ -1329,8 +1300,8 @@ void pv_display(pvstate_t state, long double esec, off_t sl, off_t tot)
 		}
 	} else {
 		if (state->control.force || pv_in_foreground()) {
-			pv_write_retry(STDERR_FILENO, state->display.display_buffer, state->display.display_string_len);
-			pv_write_retry(STDERR_FILENO, "\r", 1);
+			pv_tty_write(state, state->display.display_buffer, state->display.display_string_len);
+			pv_tty_write(state, "\r", 1);
 			state->display.display_visible = true;
 		}
 	}
