@@ -24,6 +24,32 @@
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 
+#if HAVE_MATH_H
+#include <math.h>
+#endif
+
+
+#if HAVE_SQRTL
+#else
+/*
+ * Square root of a long double.  Adapted from iputils ping/ping_common.c.
+ */
+static long ldsqrt(long double value)
+{
+	long double previous = (long double) LLONG_MAX;
+	long double result = value;
+
+	if (result > 0) {
+		while (result < previous) {
+			previous = result;
+			result = (result + (value / result)) / 2;
+		}
+	}
+
+	return result;
+}
+#endif
+
 
 /*
  * Pipe data from a list of files to standard output, giving information
@@ -34,15 +60,15 @@
 int pv_main_loop(pvstate_t state)
 {
 	long lineswritten;
-	off_t total_written, transferred_since_last, cansend;
+	off_t cansend;
 	ssize_t written;
 	long double target;
 	bool eof_in, eof_out, final_update;
 	struct timespec start_time, next_update, next_ratecheck, cur_time;
 	struct timespec init_time, next_remotecheck, transfer_elapsed;
-	long double elapsed_seconds;
-	int fd;
+	int input_fd, output_fd;
 	unsigned int file_idx;
+	bool output_is_pipe;
 
 	/*
 	 * "written" is ALWAYS bytes written by the last transfer.
@@ -50,26 +76,56 @@ int pv_main_loop(pvstate_t state)
 	 * "lineswritten" is the lines written by the last transfer,
 	 * but is only updated in line mode.
 	 *
-	 * "total_written" is the total bytes written since the start,
-	 * or in line mode, the total lines written since the start.
-	 *
-	 * "transferred_since_last" is the bytes written since the last
-	 * display, or in line mode, the lines written since the last
-	 * display.
+	 * "state->transfer.total_written" is the total bytes written since
+	 * the start, or in line mode, the total lines written since the
+	 * start.
 	 *
 	 * The remaining variables are all unchanged by linemode.
 	 */
 
-	fd = -1;
+	input_fd = -1;
+
+	output_fd = state->control.output_fd;
+	if (output_fd < 0)
+		output_fd = STDOUT_FILENO;
+
+	/* Determine whether the output is a pipe. */
+	output_is_pipe = false;
+	{
+		struct stat sb;
+		memset(&sb, 0, sizeof(sb));
+		if (0 == fstat(output_fd, &sb)) {
+			/*@-type@ */
+			if ((sb.st_mode & S_IFMT) == S_IFIFO) {
+				output_is_pipe = true;
+				debug("%s", "output is a pipe");
+			}
+			/*@+type@ *//* splint says st_mode is __mode_t, not mode_t */
+		} else {
+			debug("%s(%d): %s", "fstat", output_fd, strerror(errno));
+		}
+	}
+
+	/*
+	 * Note that we could reduce the size of the output pipe buffer like
+	 * this, to avoid the case where we write a load of data and it just
+	 * goes into the buffer so we think we're done, but the consumer
+	 * takes ages to process it:
+	 *
+	 *   fcntl(output_fd, F_SETPIPE_SZ, 4096);
+	 *
+	 * If we can peek at how much has been consumed by the other end, we
+	 * don't need to.
+	 */
 
 	pv_crs_init(state);
 
 	eof_in = false;
 	eof_out = false;
-	total_written = 0;
+	state->transfer.total_written = 0;
 	lineswritten = 0;
-	transferred_since_last = 0;
 	state->display.initial_offset = 0;
+	state->transfer.written_but_not_consumed = 0;
 
 	memset(&cur_time, 0, sizeof(cur_time));
 	memset(&start_time, 0, sizeof(start_time));
@@ -98,32 +154,31 @@ int pv_main_loop(pvstate_t state)
 	/*
 	 * Open the first readable input file.
 	 */
-	fd = -1;
-	while (fd < 0 && file_idx < state->files.file_count) {
-		fd = pv_next_file(state, file_idx, -1);
-		if (fd < 0)
+	input_fd = -1;
+	while (input_fd < 0 && file_idx < state->files.file_count) {
+		input_fd = pv_next_file(state, file_idx, -1);
+		if (input_fd < 0)
 			file_idx++;
 	}
 
 	/*
 	 * Exit early if there was no readable input file.
 	 */
-	if (fd < 0) {
+	if (input_fd < 0) {
 		if (state->control.cursor)
 			pv_crs_fini(state);
 		return state->status.exit_status;
 	}
 #if HAVE_POSIX_FADVISE
 	/* Advise the OS that we will only be reading sequentially. */
-	(void) posix_fadvise(fd, 0, 0, POSIX_FADV_SEQUENTIAL);
+	(void) posix_fadvise(input_fd, 0, 0, POSIX_FADV_SEQUENTIAL);
 #endif
 
 #ifdef O_DIRECT
 	/*
 	 * Set or clear O_DIRECT on the output.
 	 */
-	if (0 !=
-	    fcntl(STDOUT_FILENO, F_SETFL, (state->control.direct_io ? O_DIRECT : 0) | fcntl(STDOUT_FILENO, F_GETFL))) {
+	if (0 != fcntl(output_fd, F_SETFL, (state->control.direct_io ? O_DIRECT : 0) | fcntl(output_fd, F_GETFL))) {
 		debug("%s: %s", "fcntl", strerror(errno));
 	}
 	state->control.direct_io_changed = false;
@@ -137,7 +192,7 @@ int pv_main_loop(pvstate_t state)
 	if (0 == state->control.target_buffer_size) {
 		struct stat sb;
 		memset(&sb, 0, sizeof(sb));
-		if (0 == fstat(fd, &sb)) {
+		if (0 == fstat(input_fd, &sb)) {
 			size_t sz;
 			sz = (size_t) (sb.st_blksize * 32);
 			if (sz > BUFFER_SIZE_MAX) {
@@ -151,6 +206,11 @@ int pv_main_loop(pvstate_t state)
 	if (0 == state->control.target_buffer_size)
 		state->control.target_buffer_size = BUFFER_SIZE;
 
+	/*
+	 * Repeat until eof_in is true, eof_out is true, and final_update is
+	 * true.
+	 */
+
 	while ((!(eof_in && eof_out)) || (!final_update)) {
 
 		cansend = 0;
@@ -187,10 +247,10 @@ int pv_main_loop(pvstate_t state)
 		 * try to write more than we're allowed to.
 		 */
 		if ((0 < state->control.size) && (state->control.stop_at_size)) {
-			if ((state->control.size < (total_written + cansend))
+			if ((state->control.size < (state->transfer.total_written + cansend))
 			    || ((0 == cansend)
 				&& (0 == state->control.rate_limit))) {
-				cansend = state->control.size - total_written;
+				cansend = state->control.size - state->transfer.total_written;
 				if (0 >= cansend) {
 					debug("%s", "write limit reached (size explicitly set) - setting EOF flags");
 					eof_in = true;
@@ -203,7 +263,7 @@ int pv_main_loop(pvstate_t state)
 		    && (0 >= cansend) && eof_in && eof_out) {
 			written = 0;
 		} else {
-			written = pv_transfer(state, fd, &eof_in, &eof_out, cansend, &lineswritten);
+			written = pv_transfer(state, input_fd, &eof_in, &eof_out, cansend, &lineswritten);
 		}
 
 		/* End on write error. */
@@ -215,34 +275,124 @@ int pv_main_loop(pvstate_t state)
 		}
 
 		if (state->control.linemode) {
-			transferred_since_last += lineswritten;
-			total_written += lineswritten;
+			state->transfer.total_written += lineswritten;
 			if (state->control.rate_limit > 0)
 				target -= lineswritten;
 		} else {
-			transferred_since_last += written;
-			total_written += written;
+			state->transfer.total_written += written;
 			if (state->control.rate_limit > 0)
 				target -= written;
 		}
 
+#ifdef FIONREAD
+		/*
+		 * If writing to a pipe, look at how much is sitting in the
+		 * pipe buffer waiting for the receiver to read.
+		 */
+		if (output_is_pipe) {
+			int nbytes;
+			nbytes = 0;
+			if (0 != state->flag.pipe_closed) {
+				if (0 != state->transfer.written_but_not_consumed)
+					debug("%s",
+					      "clearing written_but_not_consumed because the output pipe was closed");
+				state->transfer.written_but_not_consumed = 0;
+			} else if (0 == ioctl(output_fd, FIONREAD, &nbytes)) {
+				if (nbytes >= 0) {
+					if (((size_t) nbytes) != state->transfer.written_but_not_consumed)
+						debug("%s: %d", "written_but_not_consumed is now", nbytes);
+					state->transfer.written_but_not_consumed = (size_t) nbytes;
+				} else {
+					debug("%s: %d", "FIONREAD gave a negative byte count", nbytes);
+					state->transfer.written_but_not_consumed = 0;
+				}
+			} else {
+				debug("%s(%d,%s): %s", "ioctl", output_fd, "FIONREAD", strerror(errno));
+				state->transfer.written_but_not_consumed = 0;
+			}
+		}
+#endif
+
+		state->transfer.transferred = state->transfer.total_written;
+		if (output_is_pipe && !state->control.linemode) {
+			/*
+			 * Writing bytes to a pipe - the amount transferred
+			 * to the receiver is the total amount we've
+			 * written, minus what's sitting in the pipe buffer
+			 * waiting for the receiver to consume it.
+			 */
+			state->transfer.transferred -= state->transfer.written_but_not_consumed;
+
+		} else if (output_is_pipe && state->control.linemode && state->transfer.written_but_not_consumed > 0
+			   && NULL != state->transfer.line_positions) {
+			/*
+			 * Writing lines to a pipe - similar to above, but
+			 * we have to work out how many lines the
+			 * yet-to-be-consumed data in the buffer equates to.
+			 *
+			 * To do this, we walk backwards through our record
+			 * of the line positions in the output we've
+			 * written.
+			 */
+			off_t last_consumed_position =
+			    state->transfer.last_output_position - state->transfer.written_but_not_consumed;
+			size_t lines_not_consumed = 0;
+			size_t line_from_end = 0;
+
+			/*
+			 * positions[head-1] = position of last separator written
+			 * positions[head-2] = position of second last separator written
+			 * etc
+			 *
+			 * We start at [head-1] and go backwards, wrapping
+			 * around as it's a circular buffer, stopping at the
+			 * length (number of positions stored), or when we
+			 * have gone before the last consumed position.
+			 */
+			for (line_from_end = 0; line_from_end < state->transfer.line_positions_length; line_from_end++) {
+				size_t array_index;
+				array_index =
+				    state->transfer.line_positions_head + state->transfer.line_positions_capacity -
+				    line_from_end - 1;
+				while (array_index >= state->transfer.line_positions_capacity)
+					array_index -= state->transfer.line_positions_capacity;
+				if (state->transfer.line_positions[array_index] <= last_consumed_position)
+					break;
+				lines_not_consumed++;
+			}
+
+			debug("%s: %lld -> %lld", "written_but_not_consumed bytes to lines",
+			      (unsigned long long) (state->transfer.written_but_not_consumed),
+			      (unsigned long long) lines_not_consumed);
+
+			state->transfer.transferred -= lines_not_consumed;
+		}
+
 		/*
 		 * EOF, and files remain - advance to the next file.
 		 */
 		while (eof_in && eof_out && file_idx < (state->files.file_count - 1)) {
 			file_idx++;
-			fd = pv_next_file(state, file_idx, fd);
-			if (fd >= 0) {
+			input_fd = pv_next_file(state, file_idx, input_fd);
+			if (input_fd >= 0) {
 				eof_in = false;
 				eof_out = false;
+#if HAVE_POSIX_FADVISE
+				/* Advise the OS that we will only be reading sequentially. */
+				(void) posix_fadvise(input_fd, 0, 0, POSIX_FADV_SEQUENTIAL);
+#endif
 			}
 		}
 
 		/* Now check the current time. */
 		pv_elapsedtime_read(&cur_time);
 
-		/* If full EOF, final update, and force a display updaate. */
-		if (eof_in && eof_out) {
+		/*
+		 * If we've read everything and written everything, and the
+		 * output pipe buffer is empty, then set the final update
+		 * flag, and force a display update.
+		 */
+		if (eof_in && eof_out && 0 == state->transfer.written_but_not_consumed) {
 			final_update = true;
 			if ((state->display.display_visible)
 			    || (state->control.delay_start < 0.001)) {
@@ -250,8 +400,11 @@ int pv_main_loop(pvstate_t state)
 			}
 		}
 
-		/* Just go round the loop again if there's no display. */
-		if (state->control.no_display)
+		/*
+		 * Just go round the loop again if there's no display and
+		 * we're not reporting statistics.
+		 */
+		if (state->control.no_display && !state->control.show_stats)
 			continue;
 
 		/*
@@ -320,10 +473,7 @@ int pv_main_loop(pvstate_t state)
 		memset(&transfer_elapsed, 0, sizeof(transfer_elapsed));
 		pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
 
-		elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
-
-		if (final_update)
-			transferred_since_last = -1;
+		state->transfer.elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
 
 		/* Resize the display, if a resize signal was received. */
 		if (1 == state->flag.terminal_resized) {
@@ -341,9 +491,13 @@ int pv_main_loop(pvstate_t state)
 				state->control.height = new_height;
 		}
 
-		pv_display(state, elapsed_seconds, transferred_since_last, total_written);
-
-		transferred_since_last = 0;
+		if (state->control.no_display) {
+			/* If there's no display, calculate rate for the statistics. */
+			pv_calculate_transfer_rate(state, final_update);
+		} else {
+			/* Produce the display. */
+			pv_display(state, final_update);
+		}
 	}
 
 	debug("%s: %s=%s, %s=%s", "loop ended", "eof_in", eof_in ? "true" : "false", "eof_out",
@@ -354,14 +508,60 @@ int pv_main_loop(pvstate_t state)
 	} else {
 		if ((!state->control.numeric) && (!state->control.no_display)
 		    && (state->display.display_visible))
-			pv_write_retry(STDERR_FILENO, "\n", 1);
+			pv_tty_write(state, "\n", 1);
 	}
 
 	if (1 == state->flag.trigger_exit)
 		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
 
-	if (fd >= 0)
-		(void) close(fd);
+	if (input_fd >= 0)
+		(void) close(input_fd);
+
+	/* Calculate and display the transfer statistics. */
+	if (state->control.show_stats && state->calc.measurements_taken > 0) {
+		char stats_buf[256];	 /* flawfinder: ignore */
+		long double rate_mean, rate_variance, rate_deviation;
+		int stats_size;
+
+		/* flawfinder: made safe by use of pv_snprintf() */
+
+		rate_mean = state->calc.rate_sum / ((long double) (state->calc.measurements_taken));
+		rate_variance =
+		    (state->calc.ratesquared_sum / ((long double) (state->calc.measurements_taken))) -
+		    (rate_mean * rate_mean);
+#if HAVE_SQRTL
+		rate_deviation = sqrtl(rate_variance);
+#else
+		rate_deviation = ldsqrt(rate_variance);
+#endif
+
+		debug("%s: %ld", "measurements taken", state->calc.measurements_taken);
+		debug("%s: %.3Lf", "rate_sum", state->calc.rate_sum);
+		debug("%s: %.3Lf", "ratesquared_sum", state->calc.ratesquared_sum);
+		debug("%s: %.3Lf", "rate_mean", rate_mean);
+		debug("%s: %.3Lf", "rate_variance", rate_variance);
+		debug("%s: %.3Lf", "rate_deviation", rate_deviation);
+
+		memset(stats_buf, 0, sizeof(stats_buf));
+		stats_size =
+		    pv_snprintf(stats_buf, sizeof(stats_buf), "%s = %.3Lf/%.3Lf/%.3Lf/%.3Lf %s\n",
+				_("rate min/avg/max/mdev"), state->calc.rate_min, rate_mean, state->calc.rate_max,
+				rate_deviation, state->control.bits ? _("b/s") : _("B/s"));
+
+		if (stats_size > 0 && stats_size < (int) (sizeof(stats_buf)))
+			pv_tty_write(state, stats_buf, (size_t) stats_size);
+	} else if (state->control.show_stats && state->calc.measurements_taken < 1) {
+		char msg_buf[256];	 /* flawfinder: ignore */
+		int msg_size;
+
+		/* flawfinder: made safe by use of pv_snprintf() */
+
+		memset(msg_buf, 0, sizeof(msg_buf));
+		msg_size = pv_snprintf(msg_buf, sizeof(msg_buf), "%s\n", _("rate not measured"));
+
+		if (msg_size > 0 && msg_size < (int) (sizeof(msg_buf)))
+			pv_tty_write(state, msg_buf, (size_t) msg_size);
+	}
 
 	return state->status.exit_status;
 }
@@ -377,10 +577,9 @@ int pv_main_loop(pvstate_t state)
 int pv_watchfd_loop(pvstate_t state)
 {
 	struct pvwatchfd_s info;
-	off_t position_now, total_written, transferred_since_last;
+	off_t position_now;
 	struct timespec next_update, cur_time;
 	struct timespec init_time, next_remotecheck, transfer_elapsed;
-	long double elapsed_seconds;
 	bool ended, first_check;
 	int rc;
 
@@ -427,8 +626,6 @@ int pv_watchfd_loop(pvstate_t state)
 	pv_elapsedtime_add_nsec(&next_update, (long long) (1000000000.0 * state->control.interval));
 
 	ended = false;
-	total_written = 0;
-	transferred_since_last = 0;
 	first_check = true;
 
 	while (!ended) {
@@ -448,12 +645,11 @@ int pv_watchfd_loop(pvstate_t state)
 		if (position_now < 0) {
 			ended = true;
 		} else {
-			transferred_since_last += position_now - total_written;
-			total_written = position_now;
 			if (first_check) {
 				state->display.initial_offset = position_now;
 				first_check = false;
 			}
+			state->transfer.total_written = position_now;
 		}
 
 		pv_elapsedtime_read(&cur_time);
@@ -490,10 +686,7 @@ int pv_watchfd_loop(pvstate_t state)
 		 */
 		pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
 
-		elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
-
-		if (ended)
-			transferred_since_last = -1;
+		state->transfer.elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
 
 		/* Resize the display, if a resize signal was received. */
 		if (1 == state->flag.terminal_resized) {
@@ -511,13 +704,11 @@ int pv_watchfd_loop(pvstate_t state)
 				state->control.height = new_height;
 		}
 
-		pv_display(state, elapsed_seconds, transferred_since_last, total_written);
-
-		transferred_since_last = 0;
+		pv_display(state, ended);
 	}
 
 	if (!state->control.numeric)
-		pv_write_retry(STDERR_FILENO, "\n", 1);
+		pv_tty_write(state, "\n", 1);
 
 	if (1 == state->flag.trigger_exit)
 		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
@@ -677,9 +868,8 @@ int pv_watchpid_loop(pvstate_t state)
 		displayed_lines = 0;
 
 		for (fd = 0; fd < FD_SETSIZE && NULL != info_array; fd++) {
-			off_t position_now, transferred_since_last;
+			off_t position_now;
 			struct timespec init_time, transfer_elapsed;
-			long double elapsed_seconds;
 
 			if (displayed_lines >= (int) (state->control.height))
 				break;
@@ -697,6 +887,11 @@ int pv_watchpid_loop(pvstate_t state)
 				if (pv_watchfd_changed(&(info_array[idx]))) {
 					fd_to_idx[fd] = -1;
 					info_array[idx].watch_pid = 0;
+					if (NULL != info_array[idx].state)
+						pv_state_free(info_array[idx].state);
+					/*@-mustfreeonly@ *//* not a leak - we've just free()d it. */
+					info_array[idx].state = NULL;
+					/*@+mustfreeonly@ */
 					debug("%s %d: %s", "fd", fd, "removing");
 				}
 				continue;
@@ -716,11 +911,15 @@ int pv_watchpid_loop(pvstate_t state)
 			if (position_now < 0) {
 				fd_to_idx[fd] = -1;
 				info_array[idx].watch_pid = 0;
+				if (NULL != info_array[idx].state)
+					pv_state_free(info_array[idx].state);
+				/*@-mustfreeonly@ *//* not a leak - we've just free()d it. */
+				info_array[idx].state = NULL;
+				/*@+mustfreeonly@ */
 				debug("%s %d: %s", "fd", fd, "removing");
 				continue;
 			}
 
-			transferred_since_last = position_now - info_array[idx].position;
 			info_array[idx].position = position_now;
 
 			memset(&init_time, 0, sizeof(init_time));
@@ -738,18 +937,28 @@ int pv_watchpid_loop(pvstate_t state)
 			 */
 			pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
 
-			elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
+			if (NULL != info_array[idx].state) {
+				info_array[idx].state->transfer.elapsed_seconds =
+				    pv_elapsedtime_seconds(&transfer_elapsed);
+			}
 
 			if (displayed_lines > 0) {
 				debug("%s", "adding newline");
-				pv_write_retry(STDERR_FILENO, "\n", 1);
+				pv_tty_write(state, "\n", 1);
 			}
 
-			debug("%s %d [%d]: %Lf / %Ld / %Ld", "fd", fd, idx, elapsed_seconds, transferred_since_last,
-			      position_now);
+			if (NULL == info_array[idx].state) {
+				debug("%s %d [%d]: %s / %Ld", "fd", fd, idx, "(null state)", position_now);
+			} else {
+				debug("%s %d [%d]: %Lf / %Ld", "fd", fd, idx,
+				      info_array[idx].state->transfer.elapsed_seconds, position_now);
+			}
 
-			pv_display(info_array[idx].state, elapsed_seconds, transferred_since_last, position_now);
-			displayed_lines++;
+			if (NULL != info_array[idx].state) {
+				info_array[idx].state->transfer.total_written = position_now;
+				pv_display(info_array[idx].state, false);
+				displayed_lines++;
+			}
 		}
 
 		/*
@@ -765,10 +974,10 @@ int pv_watchpid_loop(pvstate_t state)
 		while (blank_lines > 0) {
 			unsigned int x;
 			if (displayed_lines > 0)
-				pv_write_retry(STDERR_FILENO, "\n", 1);
+				pv_tty_write(state, "\n", 1);
 			for (x = 0; x < state->control.width; x++)
-				pv_write_retry(STDERR_FILENO, " ", 1);
-			pv_write_retry(STDERR_FILENO, "\r", 1);
+				pv_tty_write(state, " ", 1);
+			pv_tty_write(state, "\r", 1);
 			blank_lines--;
 			displayed_lines++;
 		}
@@ -776,7 +985,7 @@ int pv_watchpid_loop(pvstate_t state)
 		debug("%s: %d", "displayed lines", displayed_lines);
 
 		while (displayed_lines > 1) {
-			pv_write_retry(STDERR_FILENO, "\033[A", 3);
+			pv_tty_write(state, "\033[A", 3);
 			displayed_lines--;
 		}
 	}
@@ -788,14 +997,14 @@ int pv_watchpid_loop(pvstate_t state)
 	while (blank_lines > 0) {
 		unsigned int x;
 		for (x = 0; x < state->control.width; x++)
-			pv_write_retry(STDERR_FILENO, " ", 1);
-		pv_write_retry(STDERR_FILENO, "\r", 1);
+			pv_tty_write(state, " ", 1);
+		pv_tty_write(state, "\r", 1);
 		blank_lines--;
 		if (blank_lines > 0)
-			pv_write_retry(STDERR_FILENO, "\n", 1);
+			pv_tty_write(state, "\n", 1);
 	}
 	while (prev_displayed_lines > 1) {
-		pv_write_retry(STDERR_FILENO, "\033[A", 3);
+		pv_tty_write(state, "\033[A", 3);
 		prev_displayed_lines--;
 	}
 
